import json
import shutil
import time
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional

import pytest
from atomkraft.chain.testnet import VALIDATOR_DIR
from atomkraft.config.atomkraft_config import AtomkraftConfig
from atomkraft.model.traces import generate_traces
from atomkraft.utils.project import (
    ATOMKRAFT_INTERNAL_DIR,
    ATOMKRAFT_VAL_DIR_PREFIX,
    get_absolute_project_path,
    get_relative_project_path,
    project_root,
)
from caseconverter import snakecase

from ..reactor.reactor import get_reactor
from .trace import TRACE_TEST_STUB, copy_if_exists

# a key for the last used model path inside internal config
MODEL_CONFIG_KEY = "model"


def get_model() -> Path:
    """
    returns the path to the current model from the internal config
    """
    with AtomkraftConfig() as config:
        try:
            return get_absolute_project_path(config[MODEL_CONFIG_KEY])
        except KeyError:
            raise RuntimeError("Could not find any last used model.")


def test_model(
    model: Optional[Path],
    tests: List[str],
    reactor: Optional[Path],
    keypath: str,
    checker_params: Dict[str, str],
    verbose: bool,
) -> int:
    """
    Test blockchain by running one trace
    """

    if model is None:
        model = get_model()

    if reactor is None:
        reactor = get_relative_project_path(get_reactor())

    root = project_root()
    if not root:
        raise RuntimeError(
            "could not find Atomkraft project: are you in the right directory?"
        )

    timestamp = time.time()

    print(f"Generating traces for {model.name} ...")

    try:
        model_result = generate_traces(
            None, model, tests, checker_params=checker_params
        )
    except Exception as e:
        raise RuntimeError(f"[Modelator] {e}")

    test_dir = root / "tests"
    test_dir.mkdir(exist_ok=True)

    timestamp = datetime.now().isoformat(timespec="milliseconds")

    test_group = f"{model.stem}_{timestamp}"
    test_group = (
        test_group.replace("/", "_")
        .replace(".", "_")
        .replace(":", "_")
        .replace("-", "_")
    )

    test_name = f"test_{test_group}"

    successul_ops = model_result.successful()

    test_list = []

    for op in successul_ops:
        print(f"Preparing test for {op} ...")
        for trace_path in model_result.trace_paths(op):
            trace = Path(trace_path)
            if all(not c.isdigit() for c in trace.name):
                continue

            trace_name = trace.name.removesuffix(".itf.json")
            print(f"Using {trace} ...")
            trace = get_relative_project_path(trace)
            test_path = test_dir / test_name / f"test_{op}_{trace_name}.py"
            test_path.parent.mkdir(parents=True, exist_ok=True)
            with open(test_path, "w") as test:
                print(f"Writing {test_path.name} ...")
                test.write(
                    TRACE_TEST_STUB.format(
                        json.dumps(str(reactor).replace("/", ".").removesuffix(".py")),
                        json.dumps(str(trace)),
                        json.dumps(keypath),
                    )
                )
            test_list.append((trace, test_path))

    report_dir = root / "reports" / test_name
    report_dir.mkdir(parents=True, exist_ok=True)

    logging_file = report_dir / "log.txt"

    pytest_report_file = report_dir / "report.jsonl"

    pytest_args = [
        "--log-file-level=INFO",
        "--log-cli-level=INFO",
        f"--log-file={logging_file}",
        f"--report-log={pytest_report_file}",
    ]

    if verbose:
        pytest_args.append("-rP")

    val_root_dir = root / ATOMKRAFT_INTERNAL_DIR / VALIDATOR_DIR

    if val_root_dir.exists():
        shutil.rmtree(val_root_dir)

    exit_code = pytest.main(
        pytest_args + [str(test_file) for (_, test_file) in test_list]
    )

    vals_dirs = list(
        (root / ATOMKRAFT_INTERNAL_DIR / VALIDATOR_DIR).glob(
            f"{ATOMKRAFT_VAL_DIR_PREFIX}*"
        )
    )

    vals_dirs.sort(key=lambda k: k.stat().st_mtime)

    for ((trace, _), vals_dir) in zip(test_list, vals_dirs):
        copy_if_exists(
            [Path(trace), vals_dir],
            report_dir
            / snakecase(str(trace).removesuffix(".itf.json"), delimiters="./"),
        )

    if successul_ops:
        print(f"Test data is saved at {report_dir}")
    else:
        print("No trace is produced.")

    return int(exit_code)



import os
import sys
from pathlib import Path
from typing import Callable, Dict, Optional, Type

import git
import modelator
import pytest
import typer
from atomkraft import __version__
from atomkraft.utils.project import NoProjectError, project_root
from copier import run_auto

from .. import chain, test
from ..reactor.reactor import generate_reactor

GH_TEMPLATE = "gh:informalsystems/atomkraft"
GH_REVISION = "dev"


class ErrorHandlingTyper(typer.Typer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.error_handlers: Dict[Type[Exception], Callable[[Exception], int]] = {}

    def error_handler(self, exc: Type[Exception]):
        def decorator(f: Callable[[Exception], int]):
            self.error_handlers[exc] = f
            return f

        return decorator

    def __call__(self, *args, **kwargs):
        try:
            super(ErrorHandlingTyper, self).__call__(*args, **kwargs)
        except Exception as e:
            try:
                callback = self.error_handlers[type(e)]
                exit_code = callback(e)
                raise typer.Exit(code=exit_code)
            except typer.Exit as e:
                sys.exit(e.exit_code)
            except KeyError:
                print(e)
                raise typer.Exit(1)


app = ErrorHandlingTyper(
    add_completion=False,
    name="atomkraft",
    no_args_is_help=True,
    rich_markup_mode="rich",
)


def debug_callback(flag: bool):
    if not flag:
        app.pretty_exceptions_enable = False

        def exception_handler(exception_type, exception, _):
            print(f"{exception_type.__name__}: {exception}")

        sys.excepthook = exception_handler


@app.callback()
def main(
    ctx: typer.Context,
    debug: bool = typer.Option(None, callback=debug_callback),
):
    pass


@app.error_handler(NoProjectError)
def noproject_error_handler(e) -> int:
    print("You are outside of Atomkraft project")
    print("You can create an Atomkraft project using `atomkraft init <PROJECT-NAME>`")
    return 1


@app.command()
def version():
    """Print version of the atomkraft executable"""
    print(f"atomkraft {__version__}")


@app.command(
    no_args_is_help=True,
)
def init(
    name: Path = typer.Argument(..., help="Name of new directory", show_default=False),
    atomkraft_rev: Optional[str] = typer.Option(
        None, help="Atomkraft Github revision for project template"
    ),
):
    """
    Initialize new Atomkraft project in the given directory
    """
    try:
        git.Repo(os.getcwd(), search_parent_directories=True)
    except git.InvalidGitRepositoryError:
        git.Repo.init(name)
    if atomkraft_rev is None:
        atomkraft_rev = GH_REVISION
    run_auto(GH_TEMPLATE, name, vcs_ref=atomkraft_rev, data={"project_name": name})


app.add_typer(
    modelator.cli.app,
    name="model",
    short_help="Work with TLA+ models: load, typecheck, check invariants, or sample traces",
    no_args_is_help=True,
)
app.add_typer(
    chain.app,
    name="chain",
    help="Modify or control Cosmos-SDK chain",
    no_args_is_help=True,
)
app.add_typer(
    test.app,
    name="test",
    short_help="Test the blockchain based on abstract traces, either explicitly given or produced from models",
    no_args_is_help=True,
)


@app.command(
    no_args_is_help=True,
)
def smoke_test():
    """
    Run smoke tests for a Cosmos-SDK chain
    """
    pytest.main([".atomkraft/smoke_tests"])


# run with:
# atomkraft reactor --actions-list=act1,act2,act3 --variables-list=x,y,z --reactor-stub-file=path/to/the/file
# or
# atomkraft reactor --actions-list="act1, act2, act3" --variables-list="x, y, z" --reactor-stub-file="path/to/the/file"


@app.command(
    no_args_is_help=True,
)
def reactor(
    actions: str = typer.Option(
        ...,
        help="trace actions for which to create reactor stub functions",
        show_default=False,
    ),
    variables: str = typer.Option(
        ...,
        help="state variables to use as parameters for reactor stub functions.",
        show_default=False,
    ),
    path: Optional[Path] = typer.Option(
        None,
        file_okay=True,
        dir_okay=False,
        help="path where to create the reactor stub",
    ),
):
    """
    Generate a reactor stub used to interpret test traces
    """
    if path is None:
        path = project_root() / "reactors" / "reactor.py"

    actions = [act.strip() for act in actions.split(",")]
    variables = [var.strip() for var in variables.split(",")]

    if path.is_file():
        typer.confirm(
            "The stub file already exists and it will be overwritten. "
            "Are you sure you want to continue?",
            abort=True,
        )

    generate_reactor(actions, variables, stub_file_path=path)



import pytest

from ..utils.project import ATOMKRAFT_INTERNAL_DIR, project_root
from . import CHAIN_CONFIG_FILE
from .testnet import Testnet


@pytest.fixture
def testnet():
    testnet = Testnet.load_toml(
        project_root() / CHAIN_CONFIG_FILE,
        data_dir=project_root() / ATOMKRAFT_INTERNAL_DIR,
        keep=True,
        overwrite=True,
    )

    yield testnet



from pathlib import Path
from typing import Dict, List, Optional

from atomkraft.config.atomkraft_config import AtomkraftConfig
from atomkraft.config.model_config import ModelConfig
from atomkraft.utils.filesystem import last_modified_file_in
from modelator.itf import ITF
from modelator.Model import Model
from modelator.ModelResult import ModelResult


def query_configs(key: str) -> str:
    with AtomkraftConfig() as atomkraft_config:
        with ModelConfig() as model_config:
            try:
                return model_config[key] or atomkraft_config[key]
            except KeyError:
                raise FileNotFoundError


def generate_traces(
    model_config_path: Optional[Path],
    model_path: Optional[Path] = None,
    sample_operators=[],
    checker_params: Optional[Dict[str, str]] = None,
) -> ModelResult:
    """
    Call Modelator to get samples of the given model in `model_path`. Return the
    result of running the checker as a ModelResult.

    If a `model_config_path` is provided, use that to load the model path.
    Otherwise, use `model_path` and `sample_operators`.

    Raise an exception when a required parameter is absent, or when they are not
    available via defaults, or when the parameters are invalid.
    """
    init = "Init"
    next = "Next"
    traces_dir = "traces"
    with ModelConfig(model_config_path) as model_config:
        model_path = model_path or Path(model_config["model_path"])
        init = model_config.try_get("init", init)
        next = model_config.try_get("next", next)
        sample_operators = list(
            set(model_config.try_get("examples", []) + sample_operators)
        )
        traces_dir = str(Path(model_config.try_get("traces_dir", traces_dir)))

        model_config["model_path"] = str(model_path)
        model_config["init"] = init
        model_config["next"] = next
        model_config["examples"] = sample_operators
        model_config["traces_dir"] = traces_dir

    if not model_path.is_file():
        raise FileNotFoundError(f"File with model not found: {model_path}")

    if checker_params is None:
        checker_params = dict()

    model = Model.parse_file(str(model_path), init, next)
    return model.sample(
        traces_dir=traces_dir, examples=sample_operators, checker_params=checker_params
    )


def last_modified_trace_path() -> str:
    """
    Return the path to the last modified trace file in the trace directory.
    """
    try:
        traces_dir = query_configs("traces_dir")
    except FileNotFoundError:
        raise FileNotFoundError(
            "No trace directory set in neither model or configuration file."
        )

    if not Path(traces_dir).is_dir():
        raise FileNotFoundError(f"Directory does not exist: {traces_dir}")

    return last_modified_file_in(traces_dir)


def get_trace(trace_path=None) -> List["ITF"]:
    """
    Return a trace represented as an ITF class from Modelator. If `trace_path`
    is absent, return the last trace generated with the `model` command.
    """
    if not trace_path:
        trace_path = last_modified_trace_path()

    if not Path(trace_path).is_file():
        raise FileNotFoundError(f"Path {trace_path} is not a file")

    print(f"Retrieving trace from: {trace_path}")
    return ITF.from_itf_json(trace_path)



import asyncio
import socket
import tempfile
import time
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import tabulate
import tomlkit
from atomkraft.utils.project import ATOMKRAFT_INTERNAL_DIR, ATOMKRAFT_VAL_DIR_PREFIX
from grpclib.client import Channel
from terra_proto.cosmos.auth.v1beta1 import BaseAccount
from terra_proto.cosmos.auth.v1beta1 import QueryStub as AuthQueryStub
from terra_proto.cosmos.base.abci.v1beta1 import TxResponse
from terra_proto.cosmos.tx.v1beta1 import BroadcastMode, ServiceStub
from terra_sdk.client.lcd import LCDClient
from terra_sdk.client.lcd.api.tx import CreateTxOptions
from terra_sdk.core import Coins
from terra_sdk.core.fee import Fee
from terra_sdk.core.msg import Msg
from terra_sdk.key.mnemonic import MnemonicKey

from .node import Account, AccountId, Coin, ConfigPort, Node
from .utils import get_free_ports, update_port

VALIDATOR_DIR = "validator_nodes"


class Testnet:
    def __init__(
        self,
        chain_id,
        validators: Union[int, List[AccountId]],
        accounts: Union[int, List[AccountId]],
        binary: Union[str, Path],
        denom: str,
        hrp_prefix: str,
        *,
        seed: Optional[str] = None,
        coin_type: int = 118,
        config_genesis: Dict = {},
        config_node: Dict = {},
        account_balance: int = 10**27,
        validator_balance: int = 10**21,
        overwrite: bool = True,
        keep: bool = True,
        verbose: bool = False,
        data_dir: Optional[Union[str, Path]] = None,
    ):
        self.chain_id: str = chain_id
        self.set_accounts(accounts)
        self.set_validators(validators)
        self._account_seed = seed
        self.binary: Path = Path(binary) if isinstance(binary, str) else binary
        self.denom = denom
        self.hrp_prefix: str = hrp_prefix

        self.coin_type: int = coin_type
        self.config_genesis: Dict = config_genesis
        self.config_node: Dict = config_node
        self.account_balance: int = account_balance
        self.validator_balance: int = validator_balance
        self.overwrite: bool = overwrite
        self.keep = keep
        self.verbose = verbose
        if data_dir is None:
            data_dir = Path(ATOMKRAFT_INTERNAL_DIR)
        elif isinstance(data_dir, str):
            data_dir = Path(data_dir)
        self.data_dir = data_dir / VALIDATOR_DIR

    def set_accounts(self, accounts: Union[int, List[AccountId]]):
        if isinstance(accounts, list):
            self._account_ids = accounts
        elif isinstance(accounts, int):
            self._account_ids = list(range(accounts))
        else:
            self._account_ids = list(accounts)

    def set_validators(self, validators: Union[int, List[AccountId]]):
        if isinstance(validators, list):
            self._validator_ids = validators
        elif isinstance(validators, int):
            self._validator_ids = list(range(validators))
        else:
            self._validator_ids = list(validators)
        self._lead_validator = self._validator_ids[0]

    def acc_addr(self, id: AccountId) -> str:
        return self.accounts[id].address(self.hrp_prefix)

    def val_addr(self, id: AccountId, valoper: bool = False) -> str:
        if valoper:
            return self.validators[id].validator_address(self.hrp_prefix)
        else:
            return self.validators[id].address(self.hrp_prefix)

    def finalize_accounts(self):
        self.validators: Dict[AccountId, Account] = {
            v: Account(v, group="val", seed=self._account_seed)
            for v in self._validator_ids
        }
        self.accounts: Dict[AccountId, Account] = {
            a: Account(a, group="acc", seed=self._account_seed)
            for a in self._account_ids
        }

    @staticmethod
    def load_toml(path: Path, **kwargs):
        with open(path) as f:
            data = tomlkit.load(f)

        for (k, v) in kwargs.items():
            data[k] = str(v)

        return Testnet(**data)

    @staticmethod
    def ports() -> Dict[str, ConfigPort]:
        data = {}

        # config.toml
        data["p2p"] = ConfigPort("P2P", Path("config/config.toml"), "p2p.laddr")
        # data["p2p_ext"] = ConfigPort("P2P External", "config/config.toml", "p2p.external_address")
        data["abci"] = ConfigPort("ABCI", Path("config/config.toml"), "proxy_app")
        data["pprof_laddr"] = ConfigPort(
            "PPROF", Path("config/config.toml"), "rpc.pprof_laddr"
        )
        data["rpc"] = ConfigPort("RPC", Path("config/config.toml"), "rpc.laddr")
        data["prometheus"] = ConfigPort(
            "Prometheus",
            Path("config/config.toml"),
            "instrumentation.prometheus_listen_addr",
        )

        # app.toml
        data["lcd"] = ConfigPort("LCD", Path("config/app.toml"), "api.address")
        data["grpc"] = ConfigPort("gRPC", Path("config/app.toml"), "grpc.address")
        data["grpc-web"] = ConfigPort(
            "gRPC", Path("config/app.toml"), "grpc-web.address"
        )
        # dict["rosetta"] = ConfigPort("Rosetta", "config/app.toml", "rosetta.address")
        return data

    def get_validator_port(self, validator_id: AccountId, port_type: str):
        return self.validator_nodes[validator_id].get_port(self.ports()[port_type])

    def get_grpc_channel(self, validator_id: Optional[AccountId] = None) -> Channel:
        if validator_id is None:
            validator_id = self._lead_validator
        grpc_ip, grpc_port = self.get_validator_port(validator_id, "grpc").split(":", 1)
        return Channel(host=grpc_ip, port=int(grpc_port))

    def prepare(self):
        self.finalize_accounts()
        self.data_dir.mkdir(parents=True, exist_ok=True)
        self.data_dir = Path(
            tempfile.mkdtemp(prefix=ATOMKRAFT_VAL_DIR_PREFIX, dir=str(self.data_dir))
        )
        self.validator_nodes = {
            validator_id: Node(
                f"node_{validator_id}",
                self.chain_id,
                self.data_dir / f"node_{validator_id}",
                overwrite=self.overwrite,
                keep=self.keep,
                binary=Path(self.binary),
                denom=self.denom,
                hrp_prefix=self.hrp_prefix,
            )
            for validator_id in self.validators.keys()
        }

        for node in self.validator_nodes.values():
            node.init()

        for (k, v) in self.config_genesis.items():
            self.validator_nodes[self._lead_validator].set(
                Path("config/genesis.json"), v, k
            )

        for validator in self.validators.values():
            self.validator_nodes[self._lead_validator].add_account(
                Coin(self.account_balance, self.denom), validator
            )

        for account in self.accounts.values():
            self.validator_nodes[self._lead_validator].add_account(
                Coin(self.account_balance, self.denom), account
            )

        for node_id, node in self.validator_nodes.items():
            if node_id != self._lead_validator:
                self.validator_nodes[node_id].copy_genesis_from(
                    self.validator_nodes[self._lead_validator]
                )

        # very hacky
        all_ports = (
            np.array(get_free_ports(len(self.ports()) * (len(self.validators) - 1)))
            .reshape((-1, len(self.ports())))
            .tolist()
        )

        all_port_data = []

        for (node_id, node) in self.validator_nodes.items():
            for (config_file, configs) in self.config_node.items():
                for (k, v) in configs.items():
                    if isinstance(v, str):
                        # TODO: avoid tomlkit.items.str being a list
                        node.set(Path(f"config/{config_file}.toml"), str(v), k)
                    else:
                        node.set(Path(f"config/{config_file}.toml"), v, k)

            if node_id != self._lead_validator:
                ports = all_ports.pop()
                for (j, e_port) in enumerate(self.ports().values()):
                    node.update(
                        e_port.config_file,
                        lambda x: update_port(x, ports[j]),
                        e_port.property_path,
                    )

            port_data = [node.moniker]
            for e_port in self.ports().values():
                port_data.append(node.get(e_port.config_file, e_port.property_path))
            all_port_data.append(port_data)

        if self.verbose:
            print(
                tabulate.tabulate(
                    all_port_data,
                    headers=["Moniker"] + [e.title for e in self.ports().values()],
                )
            )

            print(
                tabulate.tabulate(
                    [
                        [validator.address(self.hrp_prefix)]
                        for validator in self.validators.values()
                    ],
                    headers=["Validators"],
                )
            )

            print(
                tabulate.tabulate(
                    [
                        [account.address(self.hrp_prefix)]
                        for account in self.accounts.values()
                    ],
                    headers=["Accounts"],
                )
            )

        for (node_id, node) in self.validator_nodes.items():
            node.add_key(self.validators[node_id])
            p2p = self.ports()["p2p"]
            node.add_validator(
                Coin(self.validator_balance, self.denom), self.validators[node_id]
            )

            if node_id != self._lead_validator:
                # because this
                # https://github.com/cosmos/cosmos-sdk/blob/88ee7fb2e9303f43c52bd32410901841cad491fb/x/staking/client/cli/tx.go#L599
                gentx_file = next(node.home_dir.glob("config/gentx/*json"))
                gentx_file = gentx_file.relative_to(node.home_dir)
                node_p2p = node.get(p2p.config_file, p2p.property_path).rsplit(
                    ":", maxsplit=1
                )[-1]
                node.update(gentx_file, lambda x: update_port(x, node_p2p), "body.memo")
                node.sign(self.validators[node_id], node.home_dir / gentx_file)

        for (id_a, node_a) in self.validator_nodes.items():
            for (id_b, node_b) in self.validator_nodes.items():
                if id_a != id_b:
                    node_a.copy_gentx_from(node_b)

        for node in self.validator_nodes.values():
            node.collect_gentx()

    def spinup(self):
        for node in self.validator_nodes.values():
            node.start()
        for node in self.validator_nodes.values():
            while True:
                try:
                    addr = node.get(
                        self.ports()["rpc"].config_file,
                        self.ports()["rpc"].property_path,
                    )
                    ip, port = addr.split("//")[-1].split(":")
                    with socket.create_connection(
                        (ip, port),
                        timeout=5,
                    ):
                        break
                except OSError:
                    # todo: remove this sleep in future
                    time.sleep(0.1)
                    pass

    def oneshot(self):
        self.prepare()
        self.spinup()

    def teardown(self):
        for node in self.validator_nodes.values():
            node.close()

    def broadcast_transaction(
        self,
        account_id: AccountId,
        msgs: Union[Msg, List[Msg]],
        *,
        gas: int,
        fee_amount: int,
        validator_id: Optional[AccountId] = None,
    ) -> TxResponse:
        if validator_id is None:
            validator_id = self._lead_validator

        if not isinstance(msgs, list):
            msgs = [msgs]

        account = self.accounts[account_id]

        lcdclient = LCDClient("ip", chain_id="phoenix-1")
        lcdclient.chain_id = self.chain_id

        if account.mnemonic is None:
            raise ValueError(f"Account({account.name}) do not have mnemonic")

        wallet = lcdclient.wallet(
            MnemonicKey(
                mnemonic=account.mnemonic,
                coin_type=self.coin_type,
            )
        )

        channel = self.get_grpc_channel(validator_id=validator_id)

        stub = AuthQueryStub(channel)
        result = asyncio.run(stub.account(address=account.address(self.hrp_prefix)))
        account_info = BaseAccount().parse(result.account.value)
        account_number = account_info.account_number
        sequence = account_info.sequence

        tx = wallet.create_and_sign_tx(
            CreateTxOptions(
                msgs,
                fee=Fee(gas, Coins(f"{fee_amount}{self.denom}")),
                account_number=account_number,
                sequence=sequence,
            )
        )

        service = ServiceStub(channel)

        # # BROADCAST_MODE_BLOCK defines a tx broadcasting mode where the client waits
        # # for the tx to be committed in a block.
        # BROADCAST_MODE_BLOCK = 1
        # # BROADCAST_MODE_SYNC defines a tx broadcasting mode where the client waits
        # # for a CheckTx execution response only.
        # BROADCAST_MODE_SYNC = 2
        # # BROADCAST_MODE_ASYNC defines a tx broadcasting mode where the client
        # # returns immediately.
        # BROADCAST_MODE_ASYNC = 3
        result = asyncio.run(
            service.broadcast_tx(
                tx_bytes=bytes(tx.to_proto()), mode=BroadcastMode.BROADCAST_MODE_BLOCK
            )
        ).tx_response

        channel.close()

        return result



import os
from pathlib import Path
from typing import Union

ATOMKRAFT_INTERNAL_DIR = ".atomkraft"

ATOMKRAFT_VAL_DIR_PREFIX = "val_"


class NoProjectError(RuntimeError):
    def __init__(self) -> None:
        super().__init__("Outside of Atomkraft project")


def project_root() -> Path:
    cwd = Path(os.getcwd())
    while cwd != cwd.parent:
        if (
            (cwd / "pyproject.toml").exists()
            and (cwd / "atomkraft.toml").exists()
            and (cwd / ATOMKRAFT_INTERNAL_DIR / "config.toml").exists()
        ):
            return cwd
        cwd = cwd.parent
    raise NoProjectError


def get_absolute_project_path(path: Union[Path, str]) -> Path:
    if isinstance(path, str):
        return project_root() / Path(path)
    else:
        return path.absolute()


def get_relative_project_path(path: Path) -> Path:
    if path.is_absolute():
        return path.relative_to(project_root())
    else:
        return path.absolute().relative_to(project_root())



import json
import os.path
import shutil
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Union

import pytest
from atomkraft.chain.testnet import VALIDATOR_DIR
from atomkraft.config.atomkraft_config import AtomkraftConfig
from atomkraft.utils.project import (
    ATOMKRAFT_INTERNAL_DIR,
    ATOMKRAFT_VAL_DIR_PREFIX,
    get_absolute_project_path,
    get_relative_project_path,
    project_root,
)
from caseconverter import snakecase

from ..reactor.reactor import get_reactor

# a key for the last used trace path inside internal config
TRACE_CONFIG_KEY = "trace"

TRACE_TEST_STUB = """import logging

from modelator.pytest.decorators import itf

pytest_plugins = {0}


@itf({1}, keypath={2})
def test_trace():
    logging.info("Successfully executed trace " + {1})
"""


def get_trace() -> Path:
    """
    returns the path to the current trace from the internal config
    """
    with AtomkraftConfig() as config:
        try:
            return get_absolute_project_path(config[TRACE_CONFIG_KEY])
        except KeyError:
            raise RuntimeError("Could not find any last used trace.")


def copy_if_exists(srcs: Union[Path, List[Path]], dst_path: Path):
    if isinstance(srcs, Path):
        srcs = [srcs]
    dst_path.mkdir(parents=True, exist_ok=True)
    for src in srcs:
        if src.is_dir():
            shutil.copytree(src, dst_path / src.name)
        elif src.is_file():
            shutil.copy2(src, dst_path)
        else:
            # file does not exist
            pass


def test_trace(
    trace: Optional[Path], reactor: Optional[Path], keypath: str, verbose: bool
):
    """
    Test blockchain by running one trace
    """

    root = project_root()
    if not root:
        raise RuntimeError(
            "could not find Atomkraft project: are you in the right directory?"
        )

    if trace is None:
        trace = get_trace()

    if reactor is None:
        reactor = get_relative_project_path(get_reactor())

    tests = root / "tests"
    tests.mkdir(exist_ok=True)

    timestamp = datetime.now().isoformat(timespec="milliseconds")
    test_name = f"test_{str(trace)}_{timestamp}"
    test_name = (
        test_name.replace("/", "_")
        .replace(".", "_")
        .replace(":", "_")
        .replace("-", "_")
    )
    test_path = os.path.join(tests, f"{test_name}.py")
    with open(test_path, "w") as test:
        print(f"Writing {test_name} ...")
        test.write(
            TRACE_TEST_STUB.format(
                json.dumps(str(reactor).replace("/", ".").removesuffix(".py")),
                json.dumps(str(trace)),
                json.dumps(keypath),
            )
        )
    print(f"Executing {test_name} ...")

    report_dir = root / "reports" / test_name
    report_dir.mkdir(parents=True)

    logging_file = report_dir / "log.txt"

    pytest_report_file = report_dir / "report.jsonl"

    pytest_args = [
        "--log-file-level=INFO",
        "--log-cli-level=INFO",
        f"--log-file={logging_file}",
        f"--report-log={pytest_report_file}",
    ]

    if verbose:
        pytest_args.append("-rP")

    exit_code = pytest.main(pytest_args + [test_path])

    copy_if_exists(
        [Path(trace), root / ATOMKRAFT_INTERNAL_DIR / VALIDATOR_DIR], report_dir
    )

    print(f"Test data is saved at {report_dir}")

    return int(exit_code)


def test_trace_dir(
    trace_dir: Path, reactor: Optional[Path], keypath: str, verbose: bool
):
    """
    Test blockchain by running all available traces
    """

    root = project_root()
    if not root:
        raise RuntimeError(
            "could not find Atomkraft project: are you in the right directory?"
        )

    traces = list(trace_dir.glob("**/*.itf.json"))

    if reactor is None:
        reactor = get_relative_project_path(get_reactor())

    timestamp = datetime.now().isoformat(timespec="milliseconds")

    test_group = (
        f"all_{timestamp}".replace("/", "_")
        .replace(".", "_")
        .replace(":", "_")
        .replace("-", "_")
    )

    root_test_dir = root / "tests" / test_group

    test_list = []

    for trace in traces:
        if all(not c.isdigit() for c in trace.name):
            continue
        trace = get_relative_project_path(trace)
        trace_name = (
            str(trace)
            .replace("/", "_")
            .replace(".", "_")
            .replace(":", "_")
            .replace("-", "_")
        )
        test_file_name = f"test_{trace_name}.py"
        test_path = root_test_dir / test_file_name
        test_path.parent.mkdir(parents=True, exist_ok=True)
        with open(test_path, "w") as test:
            print(f"Writing {test_file_name} ...")
            test.write(
                TRACE_TEST_STUB.format(
                    json.dumps(str(reactor).replace("/", ".").removesuffix(".py")),
                    json.dumps(str(trace)),
                    json.dumps(keypath),
                )
            )
        test_list.append((trace, test_path))

    report_dir = root / "reports" / test_group
    report_dir.mkdir(parents=True, exist_ok=True)

    logging_file = report_dir / "log.txt"

    pytest_report_file = report_dir / "report.jsonl"

    pytest_args = [
        "--log-file-level=INFO",
        "--log-cli-level=INFO",
        f"--log-file={logging_file}",
        f"--report-log={pytest_report_file}",
    ]

    if verbose:
        pytest_args.append("-rP")

    val_root_dir = root / ATOMKRAFT_INTERNAL_DIR / VALIDATOR_DIR

    if val_root_dir.exists():
        shutil.rmtree(val_root_dir)

    exit_code = pytest.main(
        pytest_args + [str(test_file) for (_, test_file) in test_list]
    )

    vals_dirs = list((val_root_dir).glob(f"{ATOMKRAFT_VAL_DIR_PREFIX}*"))

    vals_dirs.sort(key=lambda k: k.stat().st_mtime)

    for ((trace, _), vals_dir) in zip(test_list, vals_dirs):
        copy_if_exists(
            [Path(trace), vals_dir],
            report_dir
            / snakecase(str(trace).removesuffix(".itf.json"), delimiters="./"),
        )

    if traces:
        print(f"Test data is saved at {report_dir}")
    else:
        print("No trace is present.")

    return int(exit_code)



import glob
import json
import os
import shutil
from dataclasses import dataclass
from pathlib import Path
from subprocess import PIPE, Popen
from typing import Any, Callable, Dict, Optional, Union

import bip_utils
import hdwallet
import numpy as np
import tomlkit
from hdwallet.symbols import ATOM

from .. import utils


@dataclass
class ConfigPort:
    title: str
    config_file: Path
    property_path: str


class Client:
    def __init__(self, home_dir: Path, binary: Path):
        self.home_dir = home_dir
        self.binary = binary


@dataclass
class Coin:
    denom: str
    amount: int
    expoent: int

    def __init__(self, amount: int, denom: str, exponent: int = 6):
        self.amount = amount
        self.denom = denom
        self.expoent = exponent

    def __repr__(self) -> str:
        return f"{self.amount}{self.denom}"


AccountId = Union[int, str]


class Account:
    def __init__(
        self,
        name: AccountId,
        group: Optional[str] = None,
        seed: Optional[str] = None,
        strength: int = 128,
    ):
        if strength not in [128, 160, 192, 224, 256]:
            raise ValueError(
                f"Strength should be one of the following [128, 160, 192, 224, 256], but it is not {strength}."
            )
        self.name = name
        final_seed = Path(str(self.name))
        if group:
            final_seed /= group
        if seed:
            final_seed /= str(seed)
        self.entropy = (
            np.random.default_rng(list(bytes(final_seed))).bytes(strength // 8).hex()
        )
        self.wallet = hdwallet.BIP44HDWallet(symbol=ATOM).from_entropy(
            entropy=self.entropy, language="english", passphrase=""
        )

    @property
    def mnemonic(self) -> Optional[str]:
        return self.wallet.mnemonic()

    def address(self, prefix) -> str:
        return self.bech32_address(prefix)

    def validator_address(self, prefix) -> str:
        return self.bech32_address(f"{prefix}valoper")

    def bech32_address(self, prefix) -> str:
        return bip_utils.Bech32Encoder.Encode(prefix, bytes.fromhex(self.wallet.hash()))

    def __repr__(self) -> str:
        return json.dumps(self.wallet.dumps(), indent=2)


class Node:
    def __init__(
        self,
        moniker: str,
        chain_id: str,
        home_dir: Path,
        binary: Path,
        *,
        overwrite: bool = False,
        keep: bool = False,
        denom: str = "uatom",
        hrp_prefix: str = "cosmos",
    ):
        self.moniker = moniker
        self.chain_id = chain_id
        self.home_dir = home_dir
        self.binary = binary
        self.denom = denom
        self.hrp_prefix = hrp_prefix
        self.overwrite = overwrite
        self.keep = keep
        self._popen = None
        self._stdout = None
        self._stderr = None

    def init(self) -> Dict["str", Any]:
        if self.overwrite and os.path.exists(self.home_dir):
            shutil.rmtree(self.home_dir)
        args = f"init {self.moniker} --chain-id {self.chain_id}".split()
        _, data = self._execute(args, stderr=PIPE)
        return json.loads(data.decode())

    def add_key(self, account: Account):
        argstr = (
            f"keys add {account.name} --recover --keyring-backend test --output json"
        )
        stdout, stderr = self._execute(
            argstr.split(),
            stdin=f"{account.wallet.mnemonic()}\n".encode(),
            stdout=PIPE,
            stderr=PIPE,
        )
        if stdout:
            return json.loads(stdout.decode())
        return json.loads(stderr.decode())

    def add_account(self, coin: Coin, account: Account):
        argstr = f"add-genesis-account {account.address(self.hrp_prefix)} {coin} --keyring-backend test --output json"
        self._execute(argstr.split())

    def add_validator(self, coin: Coin, account: Account):
        argstr = f"gentx {account.name} {coin} --keyring-backend test --chain-id {self.chain_id} --output json"
        self._execute(argstr.split(), stderr=PIPE)

    def collect_gentx(self):
        argstr = "collect-gentxs"
        _, data = self._execute(argstr.split(), stderr=PIPE)
        return json.loads(data.decode())

    def copy_gentx_from(self, other: "Node"):
        for file in glob.glob(f"{other.home_dir}/config/gentx/*.json"):
            shutil.copy(file, f"{self.home_dir}/config/gentx")

    def copy_genesis_from(self, other: "Node"):
        for file in glob.glob(f"{other.home_dir}/config/genesis.json"):
            shutil.copy(file, f"{self.home_dir}/config")

    def get(self, path: Path, property_path: Optional[str] = None):
        with open(self.home_dir / path, encoding="utf-8") as f:
            match path.suffixes[-1]:
                case ".json":
                    data = json.load(f)
                case ".toml":
                    data = tomlkit.load(f)
                case _:
                    raise RuntimeError(f"Unexpected file {path}")
        return utils.query(data, property_path)

    def get_port(self, port_config: ConfigPort):
        return self.get(
            port_config.config_file,
            port_config.property_path,
        )

    def set(self, path: Path, value: Any, property_path: Optional[str] = None):
        if property_path is not None:
            with open(self.home_dir / path, encoding="utf-8") as f:
                match os.path.splitext(path)[-1]:
                    case ".json":
                        main_data = json.load(f)
                    case ".toml":
                        main_data = tomlkit.load(f)
                    case _:
                        raise RuntimeError(f"Unexpected file {path}")
            main_data = utils.update(main_data, property_path, value)
        else:
            main_data = value
        with open(self.home_dir / path, "w", encoding="utf-8") as f:
            match os.path.splitext(path)[-1]:
                case ".json":
                    json.dump(main_data, f)
                case ".toml":
                    tomlkit.dump(main_data, f)
                case _:
                    raise RuntimeError(f"Unexpected file {path}")

    def update(
        self,
        path: Path,
        function: Callable[[Any], Any],
        property_path: Optional[str] = None,
    ):
        self.set(path, function(self.get(path, property_path)), property_path)

    def _execute(
        self,
        args,
        *,
        stdin: Optional[bytes] = None,
        stdout: Optional[int] = None,
        stderr: Optional[int] = None,
    ):
        final_args = f"{self.binary} --home {self.home_dir}".split() + args
        # print(" ".join(final_args))
        stdin_pipe = None if stdin is None else PIPE
        with Popen(final_args, stdin=stdin_pipe, stdout=stdout, stderr=stderr) as p:
            out, err = p.communicate(input=stdin)
            rt = p.wait()
            if rt != 0:
                raise RuntimeError(f"Non-zero return code {rt}\n{err.decode()}")
            return (out, err)

    def __enter__(self):
        return self

    def sign(self, account: Account, tx_file: Path):
        argstr = f"tx sign {tx_file} --output-document {tx_file} --chain-id {self.chain_id} --overwrite --offline --sequence 0 --account-number 0 --from {account.name} --keyring-backend test --output json"
        self._execute(argstr.split())

    def start(self):
        self._stdout = open(self.home_dir / "stdout.txt", "w", encoding="utf-8")
        self._stderr = open(self.home_dir / "stderr.txt", "w", encoding="utf-8")
        args = f"{self.binary} start --home {self.home_dir}".split()
        self._popen = Popen(args, stdout=self._stdout, stderr=self._stderr)

    def close(self):
        if self._popen:
            self._popen.terminate()
            self._stdout.close()
            self._stderr.close()
            self._popen = None
            self._stdout = None
            self._stderr = None
        if not self.keep and os.path.isdir(self.home_dir):
            shutil.rmtree(self.home_dir)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()



from pathlib import Path
from typing import Any

import tomlkit


class ConfigFile(object):
    def __init__(self, path: Path):
        self.data = {}
        self.path = path

    def __enter__(self):
        self.fd = open(self.path, "a+")
        self.fd.seek(0)
        self.data = tomlkit.load(self.fd)
        return self

    def __exit__(self, type, value, traceback):
        self.fd.truncate(0)
        tomlkit.dump(self.data, self.fd)
        self.fd.close()

    def __getitem__(self, key) -> Any:
        return self.data[key]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __delitem__(self, key):
        del self.data[key]

    def get_or_update(self, key: str, value: Any) -> Any:
        try:
            return self.data[key]
        except KeyError:
            self.data[key] = value
            return value

    def try_get(self, key: str, default: Any) -> Any:
        try:
            return self.data[key]
        except KeyError:
            return default



import os
from pathlib import Path
from typing import Optional

from atomkraft.config.config_file import ConfigFile
from atomkraft.utils.project import project_root

MODEL_CONFIG_PATH = "model.toml"


class ModelConfig(ConfigFile):
    def __init__(self, path: Optional[Path] = None):
        if not path:
            if "PROJECT_ROOT" in os.environ:
                root = Path(os.environ["PROJECT_ROOT"])
            else:
                root = project_root()
            path = root / MODEL_CONFIG_PATH
        super().__init__(path)



import asyncio
import json
import socket
from typing import Any, Callable, Dict, List, Optional

import websockets
from jsonrpcclient import responses
from jsonrpcclient.requests import request_json


def get_free_ports(n: int) -> List[int]:
    """Get free ports from OS

    Args:
        n (int): Requested number of ports

    Returns:
        List[int]: List of free port
    """
    socks = []
    ports = []
    for _ in range(n):
        sock = socket.socket()
        sock.bind(("", 0))
        ports.append(sock.getsockname()[1])
        socks.append(sock)
    for sock in socks:
        sock.close()
    return ports


def update_port(old: str, new: int) -> str:
    """Replace port in `old` with `new`

    `update_port("localhost:123", "321") = "localhost:321"`

    Args:
        old (str): Old IP with port
        new (int): New port

    Returns:
        str: New IP with port
    """
    old_port = old.split(":")[-1]
    return old.replace(old_port, str(new))


class TmEventSubscribe:
    def __init__(
        self, params: Dict[str, str], filter: Optional[Callable[[Any], bool]] = None
    ):
        """Create Tendermint event subscriber.

        `with TmEventSubscribe({"tm.event": "NewBlock"}) as subscriber:`
        will exit only after a new block is generated.

        Args:
            params (Dict[str, str]): Event parsms
            filter (Optional[Callable[[Any], bool]], optional): Filter on events. Defaults to None.
        """
        self.params: Dict[str, str] = params
        if filter:
            self.filter: Callable[[Any], bool] = filter
        else:
            self.filter = lambda x: True

    def set_filter(self, func: Callable[[Any], bool]):
        """Set event filter later.

        Args:
            func (Callable[[Any], bool]): Filter on tendermint events
        """
        self.filter = func

    async def _subscribe(self):
        self.websocket = await websockets.connect("ws://0.0.0.0:26657/websocket")
        self.query = " AND ".join(
            f"{key} = '{value}'" for (key, value) in self.params.items()
        )
        req = request_json("subscribe", params={"query": self.query})
        await self.websocket.send(req)

    def __enter__(self):
        self.loop = asyncio.new_event_loop()
        self.loop.run_until_complete(self._subscribe())
        return self

    async def _wait(self):
        while True:
            msg = responses.to_result(json.loads(await self.websocket.recv()))
            if (
                isinstance(msg, responses.Ok)
                and msg.result.get("query", None) == self.query
                and self.filter(msg.result)
            ):
                break

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.loop.run_until_complete(self._wait())
        self.loop.run_until_complete(self.websocket.close())
        self.loop.close()



from pathlib import Path
from typing import List, Optional

import typer
from atomkraft.config.atomkraft_config import AtomkraftConfig
from atomkraft.utils.project import get_relative_project_path, project_root

from .model import MODEL_CONFIG_KEY, test_model
from .trace import TRACE_CONFIG_KEY, test_trace, test_trace_dir

app = typer.Typer(rich_markup_mode="rich", add_completion=False)


def FileOption(help, default):
    return typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=False,
        show_default=False,
        help=f"{help} [grey30][default: set via [bold cyan]atomkraft {default}[/bold cyan]][/grey30]",
    )


def PathOption(help, default):
    return typer.Option(
        None,
        exists=True,
        file_okay=True,
        dir_okay=True,
        show_default=False,
        help=f"{help} [grey30][default: set via [bold cyan]atomkraft {default}[/bold cyan]][/grey30]",
    )


def RequiredFileOption(help, default):
    return typer.Option(
        ...,
        exists=True,
        file_okay=True,
        dir_okay=False,
        show_default=False,
        help=f"{help} [grey30][default: set via [bold cyan]atomkraft {default}[/bold cyan]][/grey30]",
    )


@app.command()
def trace(
    # currently, require the trace to be present.
    # later, there will be an option to pick up the last one from the model
    path: Optional[Path] = PathOption(
        "trace or directory of traces to execute", "model"
    ),
    reactor: Optional[Path] = FileOption("reactor to interpret the trace", "reactor"),
    keypath: str = typer.Option(
        "action",
        show_default=True,
        help="Path to key used as step name, extracted from ITF states",
    ),
    all_: bool = typer.Option(
        False,
        "--all",
        show_default=False,
        help="Recursively find and execute traces from default trace directory",
    ),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Output logging on console"
    ),
):
    """
    Test blockchain by running one trace
    """

    if all_ or (path is not None and path.is_dir()):
        if path is None:
            path = project_root() / "traces"
        exit_code = test_trace_dir(path, reactor, keypath, verbose)
    elif path is None or path.is_file():
        exit_code = test_trace(path, reactor, keypath, verbose)
    else:
        raise RuntimeError("--trace and --all can not be used together.")

    if path and path.is_file():
        with AtomkraftConfig() as c:
            c[TRACE_CONFIG_KEY] = str(get_relative_project_path(path))

    raise typer.Exit(exit_code)


@app.command()
def model(
    model: Optional[Path] = FileOption("model used to generate traces", "model"),
    config: Optional[Path] = FileOption("model configuration", "model"),
    test: List[str] = typer.Option(
        None,
        show_default=False,
        help="model operator(s) describing test traces. multiple can be given either comma-separated, or via separate --test options",
    ),
    reactor: Optional[Path] = FileOption("reactor to interpret the traces", "reactor"),
    keypath: str = typer.Option(
        "action",
        show_default=True,
        help="Path to key used as step name, extracted from ITF states",
    ),
    max_trace: Optional[int] = typer.Option(
        None, show_default=False, help="Maximum number of traces to generate"
    ),
    view: Optional[str] = typer.Option(
        None,
        show_default=False,
        help="View projector to generate only interesting traces",
    ),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Output logging on console"
    ),
):
    """
    Test blockchain by running multiple traces generated from a model
    """
    tests = [t.strip() for ts in test for t in ts.split(",")]

    checker_args = dict()

    if max_trace:
        checker_args["max_error"] = str(max_trace)
    if view:
        checker_args["view"] = view

    exit_code = test_model(model, tests, reactor, keypath, checker_args, verbose)

    if model:
        with AtomkraftConfig() as c:
            c[MODEL_CONFIG_KEY] = str(get_relative_project_path(model))

    raise typer.Exit(exit_code)
